unit uparseadorsupersimple;
(***  la intención es tener un parseador super-simple de expresiones
   que permitea identificar las referencias a una casilla en un Libro de cálculo.
****)
{$mode delphi}

interface

uses
  Classes, SysUtils;

Type
  TTipoToken = (
    TTK_Ninguno, // -- ninguno --
    TTK_Oper, // +, -, *, etc.
    TTK_Func, // nombre(
    TTK_Apertura, // (
    TTK_Cierre,  // )
    TTk_Array, // nombre[
    TTK_Cierre_Array, // ]
    TTK_Var,   // algun_texto
    TTK_Const_S,  // String, "texto entre comillas"
    TTK_Const_F,   // Float,  1.23
    TTK_Const_I,   // Integer, Constante Entera
    TTK_Const_B,  // Boolean, Constante Booleana  0, 1
    TTK_Const_D  // double represnta fecha como TDateTime de pascal.
    );


type
  TTokensLst = class( TList )
    procedure Free;
  end;

  TNodoToken = class
    tipo: TTipoToken;
    val: string;
    constructor Create( xTipo: TTipoToken; xVal: string );
  end;



Type
  TExprRes = record
    case tipo: char of
    'B': (val_B: boolean);
    'I': (val_I: integer);
    'F': (val_F: double);
    'S': (val_S: shortstring);
    'D': (val_D: TDateTime);
  end;


  TOperProcDef = class
    nombre: shortstring;
    tiposSalidas: shortstring;
    tiposEntradas: shortstring;
    procedure evaluar( salidas, entradas: TList ); virtual; abstract;
    constructor Create( xNombre, xTiposSalidas, xTiposEntradas: shortstring );
  end;

  TExpresion = class
    val: TExprRes;
    procedure evaluar; virtual; abstract;
    // Retorna la versión serializada de la expresión.
    function Serialize: string; virtual; abstract;
    class function Simbolo: string; virtual; abstract;// debe ser sobrescrito en cada clase descendiente
  end;

  TExprLst = class( TExpresion )
    lst: TList;
    constructor Create;
    procedure add( aExpr: TExpresion );
    procedure evaluar; virtual;
    function Serialize: string; override;
    procedure Free;
  end;

  TExprNum = class( TExpresion )
    val: extended;
    // Calcula el valor "val"
    procedure evaluar; virtual; abstract;
    // Retorna la versión serializada de la expresión.
    function Serialize: string; virtual; abstract;
  end;

  TExprStr = class( TExpresion )
    val: string;
    // Calcula el valor "val"
    procedure evaluar; virtual; abstract;
    // Retorna la versión serializada de la expresión.
    function Serialize: string; virtual; abstract;
  end;

  TConst_F = class( TExprNum )
    constructor Create( valorInicial: extended );
    procedure evaluar; override;
    function Serialize: string; override;
    class function Simbolo: string; override; // debe ser sobrescrito en cada clase descendiente
  end;

  TConst_S = class( TExprStr )
    constructor Create( valorInicial: string );
    procedure evaluar; override;
    function Serialize: string; override;
    class function Simbolo: string; override; // debe ser sobrescrito en cada clase descendiente
  end;

  TVarNum = class( TExprNum )
    id: string;
    constructor Create( id: string; valorInicial: extended );
    procedure evaluar; override;
    function Serialize: string; override;
    class function Simbolo: string; override; // debe ser sobrescrito en cada clase descendiente
  end;

  TVarStr = class( TExprStr )
    id: string;
    constructor Create( id: string; valorInicial: string );
    procedure evaluar; override;
    function Serialize: string; override;
    class function Simbolo: string; override; // debe ser sobrescrito en cada clase descendiente
  end;


  TOperProc = class( TExpresion )
    template_ptr: TOperProcDef;
    parametros: TExprLst;
    constructor Create( parametros: TExprLst );
    procedure evaluar; override;
    function Serialize: string; override;
  end;

  TEvaluadorExpresionesSimples = class
    CatalogoOperadoresBinomicos: TList;
    CatalogoFuncionesEstandar: TList;
    CatalogoFuncionesDeUsuario: TList;
    CatalogoVariables: TList;
    function tokenParser( s: String ): TTokensLst;
    function TipoTokenName( tipo: TTipoToken ): string;
    function CreateFromTokenLst( tokens: TTokensLst ): TExprLst;
    function CrearExprFunc( nombreOperFunc: shortString; parametros: TExprLst ): TExpresion;
    constructor Create;
   end;

implementation


const
  Separadores = [ #0..' ', '(', '+', '-', '*', ')', '%', '>', '=', '<', '/', '^', ',', ';', ':','[',']'];
  DigitosYPunto = ['0'..'9','.'];



constructor TOperProcDef.Create( xNombre, xTiposSalidas, xTiposEntradas: shortstring );
begin
  nombre:= xNombre;
  tiposSalidas:= xTiposSalidas;
  tiposEntradas:= xTiposEntradas;
end;


procedure TTokensLst.Free;
var
  k: integer;
  nodo: TNodoToken;
begin
  for k:= 0 to count -1 do
  begin
    nodo:= items[k];
    nodo.Free;
  end;
  inherited Free;
end;



function TEvaluadorExpresionesSimples.TipoTokenName( tipo: TTipoToken ): string;
begin
    case tipo of
    TTK_Ninguno: result:= 'Ninguno';
    TTK_Oper: result:= 'Operador';
    TTK_Func: result:= 'Función';
    TTk_Array: result:= 'Array';
    TTK_Apertura: result:= 'Apertura';
    TTK_Cierre: result:= 'Cierre';
    TTK_Cierre_Array: result:= 'Cierre_Array';
    TTK_Var: result:= 'Var';
    TTK_Const_S: result:= 'String';
    TTK_Const_F: result:= 'Float';
    TTK_Const_I: result:= 'Integer';
    TTK_Const_B: result:= 'Boolean';
    end;
end;

constructor TNodoToken.Create( xTipo: TTipoToken; xVal: string );
begin
  inherited Create;
  tipo:= xTipo;
  val:= xval;
end;

function nextToken( var s: string ): TNodoToken;
var
  i: integer;
  flg_comillaAbierta: boolean;
  flg_Escape: boolean;
  c: char;
  buscando: boolean;
  res: string;
  tipoToken: TTipoToken;

begin
  s:= trim( s );
  if length( s ) = 0 then
  begin
    result:= nil;
    exit;
  end;

  c:= s[1];
  if c in Separadores then
  begin
    res:= c;
    delete( s, 1, 1 );
    case c of
      '(': tipoToken:= TTK_Apertura;
      ')': tipoToken:= TTK_Cierre;
      ']': tipoToken:= TTK_Cierre_Array;
      else
          tipoToken:= TTK_Oper;
    end;
    result:= TNodoToken.Create( tipoToken, res );
    exit;
  end;

  // Detectamdos si empieza constante encomillada.
  if c = '"' then
  begin
     delete( s, 1, 1 );
     i:= pos( '"', s );
     if i = 0 then
       raise Exception.Create( 'Error ... comilla abierta pero no cerrada!( "'+s );
     res:= copy( s, 1, i-1 );
     delete( s, 1, i );
     s:= trim( s );
     tipoToken:= TTK_Const_S;
     result:= TNodoToken.Create( tipoToken, res );
     exit;
  end;

  // Si llegamos aquí comienza o un identificador o una constante numérica
  res:= c;
  i:= 2;
  buscando:= true;
  while ( i <= length( s ) ) and buscando do
  begin
    c:= s[i];
    if  c in Separadores then
    begin
      case c of
        '(': begin
               tipoToken:= TTK_Func;
               inc( i );
             end;
        '[': begin
               tipoToken:= TTK_Array;
               inc( i );
             end;
        else
          if res[1] in DigitosYPunto then
            if pos( '.', res ) > 0 then
              tipoToken:= TTK_Const_F
            else if  (length( res ) = 1 ) and (( res[1] = '0') or (res[1] = '1' )) then
              tipoToken:= TTK_Const_B
            else
              tipoToken:= TTK_Const_I
          else
            tipoToken:= TTK_Var;
      end;
      buscando:= false;
    end
    else
    begin
      res:= res + c;
      inc( i );
    end;
  end;

  if buscando then
  begin
    s:= '';
    if res[1] in DigitosYPunto then
      if pos( '.', res ) > 0 then
        tipoToken:= TTK_Const_F
      else if  (length( res ) = 1 ) and (( res[1] = '0') or (res[1] = '1' )) then
        tipoToken:= TTK_Const_B
      else
        tipoToken:= TTK_Const_I
    else
      tipoToken:= TTK_Var;
  end
  else
    delete( s, 1, i-1 );

  result:= TNodoToken.Create( tipoToken, res );
end;





constructor TEvaluadorExpresionesSimples.Create;
begin
  CatalogoOperadoresBinomicos:= TList.Create;
  CatalogoFuncionesEstandar:= TList.Create;
  CatalogoFuncionesDeUsuario:= TList.Create;
  CatalogoVariables:= TList.Create;
end;

function TEvaluadorExpresionesSimples.tokenParser( s: String ): TTokensLst;
var
  res: TTokensLst;
  nodo: TNodoToken;
begin
  res:= TTokensLst.Create;
  nodo:= nextToken( s );
  while nodo <> nil do
  begin
    res.add( nodo );
    nodo:= nextToken( s  );
  end;
  result:= res;
end;

(**** EXPRESIONES *******)


constructor TExprLst.Create;
begin
  inherited Create;
  lst:= TList.Create;
end;

procedure TExprLst.add( aExpr: TExpresion );
begin
  lst.add( aExpr );
end;

procedure TExprLst.evaluar;
var
  k: integer;
  aExp: TExpresion;
begin
  for k:= 0 to lst.count-1 do
  begin
    aExp:= lst[k];
    aExp.evaluar;
  end;
end;

function TExprLst.Serialize: string;
var
  s: string;
  k: integer;
  aExpr: TExpresion;
begin
  s:= '(';
  for k:= 0 to lst.count -1 do
  begin
    aExpr:= lst[k];
    if k > 0 then
       s:= s+', ';
    s:= s+ aExpr.Serialize;
  end;
  s:= s+')';
  result:= s;
end;

procedure TExprLst.Free;
var
  k: integer;
  aExp: TExpresion;
begin
  for k:= 0 to lst.count-1 do
  begin
    aExp:= lst[k];
    aExp.Free;
  end;
  inherited Free;
end;


constructor TConst_F.Create( valorInicial: extended );
begin
  inherited Create;
  self.val:= valorInicial;
end;

procedure TConst_F.evaluar;
begin
end;

function TConst_F.Serialize: string;
begin
  result:= FloatToStr( val );
end;


class function TConst_F.Simbolo: string;
begin
  result:= 'Const_F';
end;


constructor TConst_S.Create( valorInicial: string );
begin
  inherited Create;
  self.val:= valorInicial;
end;

procedure TConst_S.evaluar;
begin
end;

function TConst_S.Serialize: string;
begin
  result:= val;
end;


class function TConst_S.Simbolo: string;
begin
  result:= 'Const_S';
end;

constructor TVarNum.Create( id: string; valorInicial: extended );
begin
  inherited Create;
  self.id:= id;
  self.val:= valorInicial;
end;

procedure TVarNum.evaluar;
begin
  // nada
end;
function TVarNum.Serialize: string;
begin
  result:= floatToStr( val );
end;



class function TVarNum.Simbolo: string;
begin
  result:= 'VarNum';
end;

constructor TVarStr.Create( id: string; valorInicial: string );
begin
  inherited Create;
  self.id:= id;
  self.val:= valorInicial;
end;

procedure TVarStr.evaluar;
begin
  // nada
end;

function TVarStr.Serialize: string;
begin
  result:= val;
end;


class function TVarStr.Simbolo: string;
begin
  result:= 'VarStr';
end;


function TEvaluadorExpresionesSimples.CrearExprFunc( nombreOperFunc: shortString; parametros: TExprLst ): TExpresion;
begin
  // aquí tendría que buscar en el catalogo de funciones/operadores y llamar el constructor.
  // Por ahora va un generico.
  result:= TOperProc.Create( parametros );
end;

function TEvaluadorExpresionesSimples.CreateFromTokenLst( tokens: TTokensLst ): TExprLst;
var
  aToken: TNodoToken;
  j: integer;
  sublst: TTokensLst;
  a, b: TExpresion;
  cntAperturas: integer;
  nombreOperFunc: string;
  aExprLst: TExprLst;
  waitFor: set of TTipoToken;
  buscando: boolean;
  parametros: TExprLst;

begin
  if (tokens= nil ) or (tokens.count = 0 ) then
  begin
    result:= nil;
    exit;
  end;

  aToken:= tokens[0];
  tokens.Delete( 0 );
  case aToken.tipo of
  TTK_Apertura:
    begin
      // Primero me fijo si empieza un paréntesis. Si es así
      // vamos a crear la sublista de tokens del parétnesis y cremos
      // una expresión que lo representa
      cntAperturas:= 1;
      buscando:= true;
      sublst:= TTokensLst.Create;
      while buscando and ( tokens.count > 0) do
      begin
        aToken:= tokens[0];
        case aToken.tipo of
          TTK_Cierre: dec( cntAperturas );
          TTK_Apertura, TTK_Func: inc( cntAperturas );
        end;
        if (aToken.tipo = TTK_Cierre) and (cntAperturas = 0 ) then
         buscando:= false
        else
        begin
          sublst.add( aToken );
          tokens.delete(  0 ) ;
        end;
      end;
      if buscando then
        raise Exception.create( 'Falta cerrar paréntesis' );
      a:= CreateFromTokenLst( sublst );
      waitFor:= [ TTK_Oper ];
    end;
  TTK_Func:
    begin
      // hay que leer la lista de parámetros
      NombreOperFunc:= aToken.val;
      buscando:= true;
      cntAperturas:= 1;
      buscando:= true;
      sublst:= TTokensLst.Create;
      parametros:= TExprLst.Create;
      while buscando and ( tokens.count > 0 ) do
      begin
        aToken:= tokens[0];
        tokens.delete(  0 ) ;
        case aToken.tipo of
          TTK_Cierre: dec( cntAperturas );
          TTK_Apertura, TTK_Func: inc( cntAperturas );
        end;
        if (aToken.tipo = TTK_Cierre) and (cntAperturas = 0 ) then
         buscando:= false
        else
        begin
          if (aToken.val=',') and (cntAperturas = 1) then
          begin
            aExprLst:= CreateFromTokenLst( sublst );
            parametros.add( aExprLst );
            sublst.Free;
            sublst:= TTokensLst.Create;
          end
          else
            sublst.add( aToken );
        end;
      end;
      if buscando then
        raise Exception.create( 'Falta cerrar paréntesis en Func:' );
      a:= CrearExprFunc( NombreOperFunc, parametros );
      waitFor:= [ TTK_Oper ];
    end;
  TTK_Const_F:
    begin
      a:= TConst_F.Create( StrToFloat( aToken.val ) );
      waitFor:= [ TTK_Oper ];
    end;
 TTK_Const_S:
    begin
      a:= TConst_S.Create( aToken.val );
      waitFor:= [ TTK_Oper ];
    end;
  TTK_Var:
    begin
      a:= nil;
      waitFor:= [ TTK_Oper ];
    end;
  end;

  // bien ahora tiene que venir un operador
  aToken:= tokens[0];
  if not ( aToken.tipo in WaitFor ) then
    raise Exception.Create('OOJO;... me falta temrinar de implementar');
end;


constructor TOperProc.Create( parametros: TExprLst );
begin
  inherited Create;
  self.parametros:= parametros;
end;

procedure TOperProc.evaluar;
begin
  parametros.Evaluar;
end;

function TOperProc.Serialize: string;
begin
  result:= template_ptr.nombre+'('+parametros.Serialize+')';
end;

end.
